#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#define cutoff_distance_2 1024


/* 
 * Supercon2001 sample program 
 */

typedef struct my_particle {
    double m;
    double x, y;
    double vx, vy;
    double ax, ay;
} particle;

double DT;
int n_steps;
int n_particles;
particle *particles;

void supercon2k1_get_config(int problem_num, int * size_p, 
			    int * n_steps_p, double * dt_p);
void supercon2k1_get_data(int problem_num, double * a);
void supercon2k1_verify(int problem_num, double * a);
void do_step();
void dump_mean_value(particle *);
void calc_gravity(double, double, double, double, double, double *, double *);


double round(double x){ 
  if (x < 0.0) {
    return - floor (-x + 0.5);
  } else {
    return floor (x + 0.5);
  }
} 

void calc_gravity(double xi, double yi,
		  double xj, double yj, double mj,
		  double *ax, double *ay)
{
  double dx, dy;
  double norm2;
  double cos, sin;
  double acc;

  dx = xj - xi;
  dy = yj - yi;

   if(dx > 0) dx = ceil(dx); else dx = floor(dx); 
   if(dy > 0) dy = ceil(dy); else dy = floor(dy); 

  norm2 = dx * dx + dy * dy;
  assert(norm2);
  if(norm2 < cutoff_distance_2) {
    cos = dx / sqrt(norm2);
    sin = dy / sqrt(norm2);
    acc = mj / norm2;
    *ax = round(acc * cos);
    *ay = round(acc * sin);
  } else {
    *ax = 0.0;
    *ay = 0.0;
  }
}

int main(int argc, char ** argv)
{
  int problem_num, i, j;
  double *a,*ap;
  particle *p;

  if(argc < 2){
    fprintf(stderr,"bad args...\n");
    exit(1);
  }
  problem_num = atoi(argv[1]);
    
  supercon2k1_get_config(problem_num,&n_particles,&n_steps,&DT);

  if(argc == 3){
    n_steps = atoi(argv[2]);
  }

  printf("problem_num=%d: n_particle=%d, n_steps=%d, DT=%g\n",
	 problem_num, n_particles, n_steps, DT);

  a = (double *)malloc(sizeof(double) * 5 * n_particles);
  particles = (particle *)malloc(sizeof(particle) * n_particles);
  if(a == NULL || particles == NULL){
    fprintf(stderr,"no memory...\n");
    exit(1);
  }

  supercon2k1_get_data(problem_num,a);

  ap = a;
  for(i = 0; i < n_particles; i++) {
    p = &particles[i];
    p->m = *ap++;
    p->x = *ap++;
    p->y = *ap++;
    p->vx = *ap++;
    p->vy = *ap++;
  }
for(i = 0; i < n_steps; i++){
    printf("%d\n", i);
    fflush(stdout);
    do_step();
    if(i == n_steps - 1){
	for(j = 0; j < n_particles; j++){
		printf("%d %f %f\n", j, particles[j].x, particles[j].y);
	}
	}
    }
    
  ap = a;
  for(i = 0; i < n_particles; i++) {
    p = &particles[i];
    *ap++ = p->m;
    *ap++ = p->x;
    *ap++ = p->y;
    *ap++ = p->vx;
    *ap++ = p->vy;
  }
/*
  printf("double correct_data_0_x [] = {\n");
  for(i = 0; i < n_particles; i++) 
    printf("%.10e,\n", particles[i].x);
  printf("};\n");
  
  printf("double correct_data_0_y [] = {\n");
  for(i = 0; i < n_particles; i++) 
    printf("%.10e,\n", particles[i].y); 
  printf("};\n");

  
  printf("double correct_data_0_vx [] = {\n");
  for(i = 0; i < n_particles; i++) 
    printf("%.10e,\n", particles[i].vx);
  printf("};\n");
  
  printf("double correct_data_0_vy [] = {\n");
  for(i = 0; i < n_particles; i++) 
    printf("%.10e,\n", particles[i].vy); 
  printf("};\n");
*/
  supercon2k1_verify(problem_num, (double *)a);

  return 0;
}

void dump_mean_value(particle *particles)
{
  int i;
  double mean_m, mean_v, mean_x, mean_y, mean_a;

  mean_m = mean_v = mean_x = mean_y = mean_a = 0.0; 
  for(i = 0; i < n_particles; i++) {
    double vx, vy;
    double ax, ay;
    mean_m += particles->m;
    mean_x += particles->x;
    mean_y += particles->y;
    vx = particles->vx;
    vy = particles->vy;
    mean_v += sqrt(vx * vx + vy * vy);
    ax = particles->ax;
    ay = particles->ay;
    mean_a += sqrt(ax * ax + ay * ay);
    particles++;
  }

  mean_m /= n_particles;
  mean_v /= n_particles;
  mean_x /= n_particles;
  mean_y /= n_particles;
  mean_a /= n_particles;

  printf("mass: %g, x: %g, y: %g, v: %g, a: %g\n",
	 mean_m, mean_x, mean_y, mean_v, mean_a);
}


void do_step()
{
  int i, j;
  double ax, ay;
  particle *p,*q;

  for(i = 0; i < n_particles; i++) {
    p = &particles[i];
    p->ax = 0.0;
    p->ay = 0.0;
    for(j = 0; j < n_particles; j++){
      if(i == j) continue;
      q = &particles[j];
      calc_gravity(p->x, p->y, q->x, q->y, q->m, &ax, &ay);
      p->ax += ax;
      p->ay += ay;
    }
  }

  for(i = 0; i < n_particles; i++){
    p = &particles[i];
    p->x += p->vx * DT;
    p->y += p->vy * DT;
    p->vx += p->ax * DT;
    p->vy += p->ay * DT;
  }
}

